import math
from collections import deque

import gymnasium as gym
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable


class Agent(nn.Module):
    def __init__(self, env, h_size=16):
        super(Agent, self).__init__()
        self.env = env
        # State, hidden layer, action sizes
        self.s_size = env.observation_space.shape[0]
        self.h_size = h_size
        self.a_size = env.action_space.shape[0]

        self.fc1 = nn.Linear(self.s_size, self.h_size)
        self.fc2 = nn.Linear(self.h_size, self.a_size)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = F.tanh(self.fc2(x))
        return x.cpu().data

    def set_weights(self, weights):
        """Sets the weights of the network from a numpy array."""
        s_size = self.s_size
        h_size = self.h_size
        a_size = self.a_size
        # Separate the weights for each layer
        fc1_end = (s_size * h_size) + h_size
        fc1_W = torch.from_numpy(weights[:s_size * h_size].reshape(s_size, h_size))
        fc1_b = torch.from_numpy(weights[s_size * h_size:fc1_end])
        fc2_W = torch.from_numpy(weights[fc1_end:fc1_end + (h_size * a_size)].reshape(h_size, a_size))
        fc2_b = torch.from_numpy(weights[fc1_end + (h_size * a_size):])
        # Set the weights for each layer
        self.fc1.weight.data.copy_(fc1_W.view_as(self.fc1.weight.data))
        self.fc1.bias.data.copy_(fc1_b.view_as(self.fc1.bias.data))
        self.fc2.weight.data.copy_(fc2_W.view_as(self.fc2.weight.data))
        self.fc2.bias.data.copy_(fc2_b.view_as(self.fc2.bias.data))

    def get_weights_dim(self):
        return (self.s_size + 1) * self.h_size + (self.h_size + 1) * self.a_size

    def evaluate(self, weights, gamma=1.0, max_t=5000):
        self.set_weights(weights)
        episode_return = 0.0
        state, _ = self.env.reset()
        for t in range(max_t):
            state = torch.from_numpy(state).float().to(device)
            action = self.forward(state)
            state, reward, done, _, _ = self.env.step(action)
            episode_return += reward * math.pow(gamma, t)
            if done:
                break
        return episode_return


def cem(agent, n_iterations=500, max_t=1000, gamma=1.0, print_every=5, pop_size=50, elite_frac=0.2, sigma=0.5):
    """Implementation of the cross-entropy method.

    Params
    ======
        n_iterations (int): maximum number of training iterations
        max_t (int): maximum number of timesteps per episode
        gamma (float): discount rate
        print_every (int): how often to print average score (over last 100 episodes)
        pop_size (int): size of population at each iteration
        elite_frac (float): percentage of top performers to use in update
        sigma (float): standard deviation of additive noise
    """
    n_elite = int(pop_size * elite_frac)

    scores_deque = deque(maxlen=100)
    scores = []

    # TODO 1)
    # Start with a randomly drawn mean vector ~ N(0, 1) for the network weights
    weights_dim = agent.get_weights_dim()
    weights_mean = None  # TODO

    for i_iteration in tqdm(range(1, n_iterations + 1)):
        # TODO 2)
        # Draw `pop_size` weights from the current distribution
        # Evaluate them on the environment and buffer their scores for later
        weights_pop = None
        eps_returns = None

        # TODO 3)
        # Fit the Gaussian distribution to the elite set (i.e., recalculate the distribution mean).
        # Even though we could also calculate the std. dev. on our MC samples, we keep `sigma` fixed throughout training
        elite_idxs = eps_returns.argsort()[-n_elite:]
        elite_weights = None
        weights_mean = None

        reward = agent.evaluate(weights_mean, gamma=1.0)
        scores_deque.append(reward)
        scores.append(reward)

        torch.save(agent.state_dict(), 'checkpoint.pth')

        if i_iteration % print_every == 0:
            print('Episode {}\tAverage Score: {:.2f}'.format(i_iteration, np.mean(scores_deque)))

        if np.mean(scores_deque) >= 90.0:
            print('\nEnvironment solved in {:d} iterations!\tAverage Score: {:.2f}'.format(i_iteration - 100, np.mean(scores_deque)))
            break
    return scores


if __name__ == "__main__":
    """Training"""
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    env = gym.make('MountainCarContinuous-v0')

    agent = Agent(env).to(device)
    scores = cem(agent)

    env.close()
    del agent

    # Plot the scores
    fig = plt.figure()
    ax = fig.add_subplot(111)
    plt.plot(np.arange(1, len(scores) + 1), scores)
    plt.ylabel('Score')
    plt.xlabel('Episode #')
    plt.show()

    """Rendered evaluation"""
    env = gym.make('MountainCarContinuous-v0', render_mode="human")
    agent = Agent(env).to(device)
    agent.load_state_dict(torch.load('checkpoint.pth'))

    state, _ = env.reset()
    while True:
        state = torch.from_numpy(state).float().to(device)
        with torch.no_grad():
            action = agent(state)
        env.render()
        next_state, reward, done, _, _ = env.step(action)
        state = next_state
        if done:
            break

    env.close()
